from pathlib import Path
import urllib.request
from typing import Dict, List
import json

import cv2
import numpy as np
import mediapipe as mp



MODEL_PATH = Path("assets", "models")
MODEL_DEPENDENCIES: Dict[str, List[str]] = {
    "yolov3": [
        "https://github.com/pjreddie/darknet/raw/refs/heads/master/cfg/yolov3.cfg",
        "https://github.com/patrick013/Object-Detection---Yolov3/raw/refs/heads/master/model/yolov3.weights"
    ]
}


def require_net(name: str):
    if name not in MODEL_DEPENDENCIES:
        print(f"model {name} not found")
        exit(1)

    print(f"preparing {name}")
    MODEL_PATH.mkdir(exist_ok=True, parents=True)

    for dep_url in MODEL_DEPENDENCIES[name]:
        dep_path = MODEL_PATH / dep_url.split("/")[-1]
        
        if dep_path.exists():
            continue
        
        print(f"downloading {dep_url}")
        urllib.request.urlretrieve(
            url=dep_url,
            filename=str(dep_path)
        )


def detect_humans_mediapipe(to_detect: str):
    _p = Path(to_detect)
    detected = str(_p.with_name(_p.stem + "_detected" + _p.suffix))
    boxes_file = str(_p.with_name(_p.stem + "_boxes.json"))
    print(f"detecting humans: {to_detect} => {detected}")

    # Initialize MediaPipe Pose
    mp_pose = mp.solutions.pose
    pose = mp_pose.Pose(min_detection_confidence=0.5, min_tracking_confidence=0.5)
    mp_drawing = mp.solutions.drawing_utils

    # Load your image
    image = cv2.imread(to_detect)

    # Convert the BGR image to RGB (MediaPipe uses RGB format)
    image_rgb = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

    # Perform pose detection
    results = pose.process(image_rgb)

    # Draw landmarks on the image (optional)
    if results.pose_landmarks:
        mp_drawing.draw_landmarks(image, results.pose_landmarks, mp_pose.POSE_CONNECTIONS)

        # You can extract specific keypoints (head, feet, etc.)
        # Example: 0 = Nose, 15 = Left Foot, 16 = Right Foot
        head = results.pose_landmarks.landmark[0]  # Nose landmark
        left_foot = results.pose_landmarks.landmark[15]  # Left Foot
        right_foot = results.pose_landmarks.landmark[16]  # Right Foot

        # Convert to pixel coordinates
        h, w, _ = image.shape
        head_coords = int(head.x * w), int(head.y * h)
        left_foot_coords = int(left_foot.x * w), int(left_foot.y * h)
        right_foot_coords = int(right_foot.x * w), int(right_foot.y * h)

        # Draw head and feet positions on the image
        cv2.circle(image, head_coords, 5, (255, 0, 0), -1)  # Head in blue
        cv2.circle(image, left_foot_coords, 5, (0, 0, 255), -1)  # Left foot in red
        cv2.circle(image, right_foot_coords, 5, (0, 0, 255), -1)  # Right foot in red

    # Save the result
    cv2.imwrite(detected, image)


def detect_humans(to_detect: str):
    _p = Path(to_detect)
    detected = str(_p.with_name(_p.stem + "_detected" + _p.suffix))
    boxes_file = str(_p.with_name(_p.stem + "_boxes.json"))
    print(f"detecting humans: {to_detect} => {detected}")

    require_net("yolov3")

    # Load YOLO
    net = cv2.dnn.readNet(str(MODEL_PATH / 'yolov3.weights'), str(MODEL_PATH / 'yolov3.cfg'))
    layer_names = net.getLayerNames()
    indices = net.getUnconnectedOutLayers()
    output_layers = [layer_names[int(i) - 1] for i in indices]


    # Load image
    image = cv2.imread(to_detect)
    original_image = cv2.imread(to_detect)
    height, width, channels = image.shape

    # Create blob and do forward pass
    blob = cv2.dnn.blobFromImage(image, 0.00392, (416, 416), (0, 0, 0), True, crop=False)
    net.setInput(blob)
    outs = net.forward(output_layers)

    boxes = []
    confidences = []

    # Information for each object detected
    for out in outs:
        for detection in out:
            scores = detection[5:]
            class_id = np.argmax(scores)
            confidence = scores[class_id]
            if confidence > 0.5 and class_id == 0:  # Class ID 0 is human
                center_x = int(detection[0] * width)
                center_y = int(detection[1] * height)
                w = int(detection[2] * width)
                h = int(detection[3] * height)
                x = int(center_x - w / 2)
                y = int(center_y - h / 2)

                boxes.append([x, y, w, h])
                confidences.append(float(confidence))

    # Apply Non-Maximum Suppression
    indices = cv2.dnn.NMSBoxes(boxes, confidences, score_threshold=0.5, nms_threshold=0.4)

    boxes_structures = {}
    human_boxes = boxes_structures["humans"] = []

    human_part_folder = _p.with_name(_p.stem + "_parts")
    human_part_folder.mkdir(exist_ok=True)

    for i in indices:
        i = i[0] if isinstance(i, (list, np.ndarray)) else i  # Flatten index if needed
        x, y, w, h = boxes[i]

        human_part_image_path = human_part_folder / (_p.stem + "_" + str(i) + _p.suffix)
        human_crop = original_image[y:y+h, x:x+w]
        cv2.imwrite(str(human_part_image_path), human_crop)

        print(f"\tfound human at {x}/{y} with the size of {w} x {h}")
        human_boxes.append({
            "x": x,
            "y": y,
            "w": w,
            "h": h,
            "cropped": str(human_part_image_path)
        })


        cv2.rectangle(image, (x, y), (x + w, y + h), (0, 255, 255), 2)


    # Save the result
    with open(boxes_file, "w") as f:
        json.dump(boxes_structures, f)
    cv2.imwrite(detected, image)
